/***
 * Juliac
 * Copyright (C) 2012-2013a, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.compile.janino;

import java.io.File;
import java.io.IOException;
import java.util.List;

import org.codehaus.commons.compiler.CompileException;
import org.codehaus.janino.ClassLoaderIClassLoader;
import org.codehaus.janino.IClassLoader;
import org.codehaus.janino.WarningHandler;
import org.codehaus.janino.util.resource.DirectoryResourceCreator;
import org.codehaus.janino.util.resource.DirectoryResourceFinder;
import org.codehaus.janino.util.resource.PathResourceFinder;
import org.codehaus.janino.util.resource.Resource;
import org.codehaus.janino.util.resource.ResourceCreator;
import org.codehaus.janino.util.resource.ResourceFinder;
import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.SourceFile;
import org.objectweb.fractal.juliac.api.CompileSupportItf;
import org.objectweb.fractal.juliac.conf.JuliacConfig;

/**
 * This class implements a compilation service based on the Janino compiler.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 2.5
 */
public class CompileSupportImpl implements CompileSupportItf {

    // -----------------------------------------------------------------------
    // Implementation of the JuliacModuleItf interface
    // -----------------------------------------------------------------------

	final public static Class<CompileSupportItf> SERVICE_TYPE =
		CompileSupportItf.class;
	
	private Juliac jc;
	
    public void init( Juliac jc ) {
    	this.jc = jc;
        JuliacConfig jconf = jc.getJuliacConfig();
        jconf.register(SERVICE_TYPE,this);
    }
    
    public void close( Juliac jc ) {
        JuliacConfig jconf = jc.getJuliacConfig();
        jconf.unregister(SERVICE_TYPE,this);
    }

    
    // -----------------------------------------------------------------------
    // Implementation of the CompileSupportItf interface
    // -----------------------------------------------------------------------

    /**
     * Compile the specified source files and dump the resulting bytecode files
     * into the specified directory.
     * 
     * @param sfs       the source files
     * @param classDir  the directory for storing bytecode files
     * @return 
     * 		the paths relative to <code>classDir</code> of compiled bytecode
     * 		files 
     * @throws IOException  if an error occurs during the compilation
     */
    public List<String> compile( List<SourceFile> sfs, File classDir )
    throws IOException {
        
    	JuliacConfig jconf = jc.getJuliacConfig();
    	ClassLoader classLoader = jconf.getClassLoader();
    	
        /*
         * Prepare compiler options.
         * 
         * TODO classFileFinder and classFileCreator can be configured to
         * compile in memory instead of in a directory. See the Janino Compiler
         * source code for examples.
         */
    	File[] classPath = { new File(".") };
    	ResourceFinder sourceFinder = new PathResourceFinder(classPath);
    	IClassLoader iClassLoader = new ClassLoaderIClassLoader(classLoader);
    	ResourceFinder classFileFinder = new DirectoryResourceFinder(classDir);
    	ResourceCreator classFileCreator =
			new DirectoryResourceCreator(classDir);
    	String optionalCharacterEncoding = null;
    	boolean verbose = false;
    	boolean debugSource = false;
    	boolean debugLines = false;
    	boolean debugVars = false;
    	WarningHandler optionalWarningHandler = new WarningHandlerImpl(jc);
	
    	JaninoCompiler compiler =
			new JaninoCompiler(
				jc,
				sourceFinder, iClassLoader, classFileFinder, classFileCreator,
				optionalCharacterEncoding, verbose, debugSource, debugLines,
				debugVars, optionalWarningHandler );
    	
        /*
         * Prepare the source files.
         */
    	Resource[] resources = new Resource[sfs.size()];
    	for (int i = 0; i < resources.length; i++) {
    		SourceFile sf = sfs.get(i);
			String qname = sf.getName();
			char[] content = sf.getContent();
			resources[i] = new CharArrayResource(qname,content);
		}
    	
        /*
         * Compile.
         */
        List<String> outputFileNames = null;
    	try {
    		outputFileNames = compiler.compile(resources);
		}
    	catch( CompileException ce ) {
    		IOException ioe = new IOException();
    		ioe.initCause(ce);
    		throw ioe;
		}

    	return outputFileNames;
    }
}
