/***
 * Juliac
 * Copyright (C) 2012-2013a, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.compile.janino;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

import org.codehaus.commons.compiler.CompileException;
import org.codehaus.commons.compiler.Location;
import org.codehaus.janino.ClassFileIClass;
import org.codehaus.janino.Descriptor;
import org.codehaus.janino.IClass;
import org.codehaus.janino.IClassLoader;
import org.codehaus.janino.JaninoRuntimeException;
import org.codehaus.janino.Java;
import org.codehaus.janino.Parser;
import org.codehaus.janino.Scanner;
import org.codehaus.janino.UnitCompiler;
import org.codehaus.janino.WarningHandler;
import org.codehaus.janino.util.ClassFile;
import org.codehaus.janino.util.resource.FileResource;
import org.codehaus.janino.util.resource.FileResourceCreator;
import org.codehaus.janino.util.resource.Resource;
import org.codehaus.janino.util.resource.ResourceCreator;
import org.codehaus.janino.util.resource.ResourceFinder;
import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.conf.JuliacConfig;

/**
 * A modified version of the Janino Compiler class for integration with Juliac.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 2.5
 */
public class JaninoCompiler {

	private Juliac jc;
	
    private final ResourceFinder        classFileFinder;
    /** Special value for "classFileResourceFinder". */
    public static final ResourceFinder  FIND_NEXT_TO_SOURCE_FILE = null;
    private final ResourceCreator       classFileCreator;
    /** Special value for "classFileResourceCreator". */
    public static final ResourceCreator CREATE_NEXT_TO_SOURCE_FILE = null;
    private final String                optionalCharacterEncoding;
    private final boolean               debugSource;
    private final boolean               debugLines;
    private final boolean               debugVars;
    private final WarningHandler        optionalWarningHandler;
    private UnitCompiler.ErrorHandler   optionalCompileErrorHandler = null;

    private final IClassLoader iClassLoader;
    private final ArrayList<UnitCompiler> parsedCompilationUnits =
    		new ArrayList<UnitCompiler>(); // UnitCompiler

    /**
     * To mimic the behavior of JAVAC with a missing "-d" command line option,
     * pass {@link #FIND_NEXT_TO_SOURCE_FILE} as the <code>classFileResourceFinder</code> and
     * {@link #CREATE_NEXT_TO_SOURCE_FILE} as the <code>classFileResourceCreator</code>.
     * <p>
     * If it is impossible to check whether an already-compiled class file
     * exists, or if you want to enforce recompilation, pass
     * {@link ResourceFinder#EMPTY_RESOURCE_FINDER} as the
     * <code>classFileResourceFinder</code>.
     *
     * @param jc           The Juliac instance this object is associated with
     * @param sourceFinder Finds extra Java compilation units that need to be compiled (a.k.a. "sourcepath")
     * @param iClassLoader loads auxiliary {@link IClass}es; e.g. <code>new ClassLoaderIClassLoader(ClassLoader)</code>
     * @param classFileFinder Where to look for up-to-date class files that need not be compiled
     * @param classFileCreator Used to store generated class files
     * @param optionalCharacterEncoding
     * @param verbose
     * @param optionalWarningHandler used to issue warnings
     */
    public JaninoCompiler(
        Juliac jc,
    	ResourceFinder  sourceFinder,
        IClassLoader    iClassLoader,
        ResourceFinder  classFileFinder,
        ResourceCreator classFileCreator,
        final String    optionalCharacterEncoding,
        boolean         verbose,
        boolean         debugSource,
        boolean         debugLines,
        boolean         debugVars,
        WarningHandler  optionalWarningHandler
    ) {
    	this.jc                        = jc;
        this.classFileFinder           = classFileFinder;
        this.classFileCreator          = classFileCreator;
        this.optionalCharacterEncoding = optionalCharacterEncoding;
        this.debugSource               = debugSource;
        this.debugLines                = debugLines;
        this.debugVars                 = debugVars;
        this.optionalWarningHandler    = optionalWarningHandler;

        // Set up the IClassLoader.
        this.iClassLoader = new CompilerIClassLoader(
            sourceFinder,
            iClassLoader
        );
    }

    /**
     * @param sourceResources Contain the compilation units to compile
     */
    public List<String> compile( Resource[] sourceResources )
	throws CompileException, IOException {

        // Set up the compile error handler as described at "setCompileErrorHandler()".
        UnitCompiler.ErrorHandler ceh = (
            this.optionalCompileErrorHandler != null
            ? this.optionalCompileErrorHandler
            : new UnitCompiler.ErrorHandler() {
                int compileErrorCount = 0;
                public void handleError( String message, Location optionalLocation )
        		throws CompileException {
                    CompileException ex =
                		new CompileException(message,optionalLocation);
                    if (++this.compileErrorCount >= 20) throw ex;
                    JuliacConfig jconf = jc.getJuliacConfig();
                    Logger logger = jconf.getLogger();
                    logger.severe(ex.getMessage());
                }
            }
        );


        // Parse all source files.
        this.parsedCompilationUnits.clear();
        for (int i = 0; i < sourceResources.length; ++i) {
            this.parsedCompilationUnits.add(new UnitCompiler(this.parseCompilationUnit(
                sourceResources[i].getFileName(),                   // fileName
                new BufferedInputStream(sourceResources[i].open()), // inputStream
                this.optionalCharacterEncoding                      // optionalCharacterEncoding
            ), this.iClassLoader));
        }

        // Compile all parsed compilation units. The vector of parsed CUs may
        // grow while they are being compiled, but eventually all CUs will
        // be compiled.
        List<String> outputFileNames = new ArrayList<String>();
        for (int i = 0; i < this.parsedCompilationUnits.size(); ++i) {
            UnitCompiler unitCompiler = parsedCompilationUnits.get(i);
            Java.CompilationUnit cu = unitCompiler.compilationUnit;
            if( cu.optionalFileName == null )
            	throw new JaninoRuntimeException();
            File sourceFile = new File(cu.optionalFileName);

            unitCompiler.setCompileErrorHandler(ceh);
            unitCompiler.setWarningHandler(this.optionalWarningHandler);

            ClassFile[] classFiles =
        		unitCompiler.compileUnit(debugSource,debugLines,debugVars);

            // Store the compiled classes and interfaces into class files.
            for (int j = 0; j < classFiles.length; ++j) {
                this.storeClassFile( classFiles[j], sourceFile );
                String outputFileName = classFiles[j].getThisClassName();
                outputFileNames.add(outputFileName);
            }
        }

        return outputFileNames;
    }

    /**
     * Read one compilation unit from a file and parse it.
     * <p>
     * The <code>inputStream</code> is closed before the method returns.
     * @return the parsed compilation unit
     */
    private Java.CompilationUnit parseCompilationUnit(
        String      fileName,
        InputStream inputStream,
        String      optionalCharacterEncoding
    ) throws CompileException, IOException {
        try {
            Scanner scanner = new Scanner(fileName, inputStream, optionalCharacterEncoding);
            scanner.setWarningHandler(this.optionalWarningHandler);
            Parser parser = new Parser(scanner);
            parser.setWarningHandler(this.optionalWarningHandler);

            return parser.parseCompilationUnit();
        } finally {
            inputStream.close();
        }
    }

    /**
     * Store the byte code of this {@link ClassFile} in the file system. Directories are created
     * as necessary.
     * @param classFile
     * @param sourceFile Required to compute class file path if no destination directory given
     */
    public void storeClassFile(ClassFile classFile, final File sourceFile) throws IOException {
        String classFileResourceName = ClassFile.getClassFileResourceName(classFile.getThisClassName());

        // Determine where to create the class file.
        ResourceCreator rc;
        if (this.classFileCreator != JaninoCompiler.CREATE_NEXT_TO_SOURCE_FILE) {
            rc = this.classFileCreator;
        } else {

            // If the JAVAC option "-d" is given, place the class file next
            // to the source file, irrespective of the package name.
            rc = new FileResourceCreator() {
                protected File getFile(String resourceName) {
                    return new File(
                        sourceFile.getParentFile(),
                        resourceName.substring(resourceName.lastIndexOf('/') + 1)
                    );
                }
            };
        }
        OutputStream os = rc.createResource(classFileResourceName);
        try {
            classFile.store(os);
        } catch (IOException ex) {
            try { os.close(); } catch (IOException e) { }
            os = null;
            if (!rc.deleteResource(classFileResourceName)) {
                throw new IOException(
                    "Could not delete incompletely written class file \""
                    + classFileResourceName
                    + "\""
                );
            }
            throw ex;
        } finally {
            if (os != null) try { os.close(); } catch (IOException e) { }
        }
    }

    /**
     * A specialized {@link IClassLoader} that loads {@link IClass}es from the following
     * sources:
     * <ol>
     *   <li>An already-parsed compilation unit
     *   <li>A class file in the output directory (if existant and younger than source file)
     *   <li>A source file in any of the source path directories
     *   <li>The parent class loader
     * </ol>
     * Notice that the {@link CompilerIClassLoader} is an inner class of {@link JaninoCompiler} and
     * heavily uses {@link JaninoCompiler}'s members.
     */
    private class CompilerIClassLoader extends IClassLoader {
        private final ResourceFinder sourceFinder;

        /**
         * @param sourceFinder Where to look for source files
         * @param optionalParentIClassLoader {@link IClassLoader} through which {@link IClass}es are to be loaded
         */
        public CompilerIClassLoader(
            ResourceFinder sourceFinder,
            IClassLoader   optionalParentIClassLoader
        ) {
            super(optionalParentIClassLoader);
            this.sourceFinder = sourceFinder;
            super.postConstruct();
        }

        /**
         * @param type field descriptor of the {@IClass} to load, e.g. "Lpkg1/pkg2/Outer$Inner;"
         * @return <code>null</code> if a the type could not be found
         * @throws ClassNotFoundException if an exception was raised while loading the {@link IClass}
         */
        protected IClass findIClass(final String type) throws ClassNotFoundException {

            // Determine the class name.
            String className = Descriptor.toClassName(type); // E.g. "pkg1.pkg2.Outer$Inner"

            // Do not attempt to load classes from package "java".
            if (className.startsWith("java.")) return null;

            // Determine the name of the top-level class.
            String topLevelClassName;
            {
                int idx = className.indexOf('$');
                topLevelClassName = idx == -1 ? className : className.substring(0, idx);
            }

            // Check the already-parsed compilation units.
            for (int i = 0; i < JaninoCompiler.this.parsedCompilationUnits.size(); ++i) {
                UnitCompiler uc = (UnitCompiler) JaninoCompiler.this.parsedCompilationUnits.get(i);
                IClass res = uc.findClass(topLevelClassName);
                if (res != null) {
                    if (!className.equals(topLevelClassName)) {
                        res = uc.findClass(className);
                        if (res == null) return null;
                    }
                    this.defineIClass(res);
                    return res;
                }
            }

            // Search source path for uncompiled class.
            final Resource sourceResource = this.sourceFinder.findResource(ClassFile.getSourceResourceName(className));
            if (sourceResource == null) return null;

            // Find an existing class file.
            Resource classFileResource;
            if (JaninoCompiler.this.classFileFinder != JaninoCompiler.FIND_NEXT_TO_SOURCE_FILE) {
                classFileResource = JaninoCompiler.this.classFileFinder.findResource(
                    ClassFile.getClassFileResourceName(className)
                );
            } else {
                if (!(sourceResource instanceof FileResource)) return null;
                File classFile = new File(
                    ((FileResource) sourceResource).getFile().getParentFile(),
                    ClassFile.getClassFileResourceName(className.substring(className.lastIndexOf('.') + 1))
                );
                classFileResource = classFile.exists() ? new FileResource(classFile) : null;
            }

            // Compare source modification time against class file modification time.
            if (classFileResource != null && sourceResource.lastModified() <= classFileResource.lastModified()) {

                // The class file is up-to-date; load it.
                return this.defineIClassFromClassFileResource(classFileResource);
            } else {

                // Source file not yet compiled or younger than class file.
                return this.defineIClassFromSourceResource(sourceResource, className);
            }
        }

        /**
         * Parse the compilation unit stored in the given <code>sourceResource</code>, remember it in
         * <code>Compiler.this.parsedCompilationUnits</code> (it may declare other classes that
         * are needed later), find the declaration of the type with the given
         * <code>className</code>, and define it in the {@link IClassLoader}.
         * <p>
         * Notice that the CU is not compiled here!
         */
        private IClass defineIClassFromSourceResource(
            Resource sourceResource,
            String                  className
        ) throws ClassNotFoundException {

            // Parse the source file.
            UnitCompiler uc;
            try {
                Java.CompilationUnit cu = JaninoCompiler.this.parseCompilationUnit(
                    sourceResource.getFileName(),                   // fileName
                    new BufferedInputStream(sourceResource.open()), // inputStream
                    JaninoCompiler.this.optionalCharacterEncoding         // optionalCharacterEncoding
                );
                uc = new UnitCompiler(cu, JaninoCompiler.this.iClassLoader);
            } catch (IOException ex) {
                throw new ClassNotFoundException("Parsing compilation unit \"" + sourceResource + "\"", ex);
            } catch (CompileException ex) {
                throw new ClassNotFoundException("Parsing compilation unit \"" + sourceResource + "\"", ex);
            }

            // Remember compilation unit for later compilation.
            JaninoCompiler.this.parsedCompilationUnits.add(uc);

            // Define the class.
            IClass res = uc.findClass(className);
            if (res == null) {

                // This is a really complicated case: We may find a source file on the source
                // path that seemingly contains the declaration of the class we are looking
                // for, but doesn't. This is possible if the underlying file system has
                // case-insensitive file names and/or file names that are limited in length
                // (e.g. DOS 8.3).
                return null;
            }
            this.defineIClass(res);
            return res;
        }

        /**
         * Open the given <code>classFileResource</code>, read its contents, define it in the
         * {@link IClassLoader}, and resolve it (this step may involve loading more classes).
         */
        private IClass defineIClassFromClassFileResource(Resource classFileResource) throws ClassNotFoundException {
            InputStream is = null;
            ClassFile cf;
            try {
                is = classFileResource.open();
                cf = new ClassFile(new BufferedInputStream(is));
            } catch (IOException ex) {
                throw new ClassNotFoundException("Opening class file resource \"" + classFileResource + "\"", ex);
            } finally {
                if (is != null) try { is.close(); } catch (IOException e) { }
            }
            ClassFileIClass result = new ClassFileIClass(
                cf,                       // classFile
                CompilerIClassLoader.this // iClassLoader
            );

            // Important: We must FIRST call "defineIClass()" so that the
            // new IClass is known to the IClassLoader, and THEN
            // "resolveAllClasses()", because otherwise endless recursion could
            // occur.
            this.defineIClass(result);
            result.resolveAllClasses();

            return result;
        }
    }
}


